#include "boxm2_multi_store_aux.h"
//:
// \file

#include <boxm2_multi_util.h>

//ocl includes
#include <bocl/bocl_manager.h>
#include <bocl/bocl_mem.h>
#include <bocl/bocl_device.h>
#include <bocl/bocl_kernel.h>

#include <vcl_algorithm.h>
#include <boxm2/boxm2_scene.h>
#include <boxm2/boxm2_util.h>
#include <boxm2/ocl/boxm2_ocl_util.h>
#include <boxm2/ocl/boxm2_opencl_cache.h>
#include <boxm2/ocl/algo/boxm2_ocl_camera_converter.h>

#include <vil/vil_image_view_base.h>
#include <vil/vil_save.h>
#include <vpgl/vpgl_perspective_camera.h>
#include <vul/vul_timer.h>

vcl_map<vcl_string, bocl_kernel*> boxm2_multi_store_aux::kernels_;

//-------------------------------------------------------------
// Stores seg len and observation in cell-level aux data
//-------------------------------------------------------------
float boxm2_multi_store_aux::store_aux(boxm2_multi_cache&       cache,
                                       vil_image_view<float>&   img,
                                       vpgl_camera_double_sptr  cam,
                                       boxm2_multi_update_helper& helper)
{
  vcl_cout<<"  -- boxm2_multi_store_aux store aux --"<<vcl_endl;
  //verify appearance model
  vcl_size_t lthreads[2] = {8,8};
  vcl_string data_type, options;
  int apptypesize;
  if ( !boxm2_multi_util::get_scene_appearances(cache.get_scene(), data_type, options, apptypesize) )
    return 0.0f;

  //setup image size
  int ni=img.ni(),
      nj=img.nj();
  unsigned cl_ni=RoundUp(ni,lthreads[0]);
  unsigned cl_nj=RoundUp(nj,lthreads[1]);
  vcl_size_t gThreads[] = {cl_ni,cl_nj};

  //input image buffer
  float* inImg;
  if(img.size() == cl_ni*cl_nj){
    inImg = img.top_left_ptr();
  }
  else {
    inImg = new float[cl_ni*cl_nj];
    int c = 0;
    for (int j=0; j<nj; ++j)
      for (int i=0; i<ni; ++i)
        inImg[c++] = img(i,j);
  }
  //-------------------------------------------------------
  //prepare buffers for each device
  //-------------------------------------------------------
  vcl_vector<bocl_mem_sptr> in_imgs;
  vcl_vector<cl_command_queue>& queues = helper.queues_;
  vcl_vector<bocl_mem_sptr>& out_imgs = helper.outputs_,
                             img_dims = helper.img_dims_,
                             ray_ds = helper.ray_ds_,
                             ray_os = helper.ray_os_,
                             lookups = helper.lookups_;
  vcl_size_t maxBlocks = helper.maxBlocks_;
  vcl_vector<boxm2_opencl_cache*>& ocl_caches = helper.vis_caches_;
  for (int i=0; i<ocl_caches.size(); ++i) {
    //grab sub scene and it's cache
    boxm2_opencl_cache* ocl_cache = ocl_caches[i];
    boxm2_scene_sptr    sub_scene = ocl_cache->get_scene();
    bocl_device_sptr    device    = ocl_cache->get_device();

    //create image var
    bocl_mem_sptr in_mem = ocl_cache->alloc_mem(cl_ni*cl_nj*sizeof(float),inImg,"exp image buffer");
    in_mem->create_buffer(CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR);
    in_imgs.push_back(in_mem);
  }


  //----------------------------------------------------------------
  // Call per block/per scene update (to ensure cpu-> gpu cache works
  //---------------------------------------------------------------
  //visibility order
  vcl_vector<boxm2_multi_cache_group*> grp = helper.group_orders_; //cache.get_vis_groups(cam);
  vcl_cout<<"Group list size; "<<grp.size()<<vcl_endl;
  vul_timer t; t.mark();
  float transfer_time = 0.0f;
  for(int grpId=0; grpId<grp.size(); ++grpId) {
    boxm2_multi_cache_group& group = *grp[grpId];
    vcl_vector<boxm2_block_id>& ids = group.ids();
    vcl_vector<int> indices = group.order_from_cam(cam);
    for(int idx=0; idx<indices.size(); ++idx) {
      int i = indices[idx];
      //grab sub scene and it's cache
      boxm2_opencl_cache* ocl_cache = ocl_caches[i];
      boxm2_scene_sptr    sub_scene = ocl_cache->get_scene();
      bocl_device_sptr    device    = ocl_cache->get_device();

      // compile the kernel/retrieve cached kernel for this device
      bocl_kernel* kern = get_kernels(device, options);

      //Run block store aux
      boxm2_block_id id = ids[i];
      store_aux_per_block(id, sub_scene, ocl_cache, queues[i], kern,
                          in_imgs[i], img_dims[i], ray_os[i], ray_ds[i],
                          out_imgs[i], lookups[i], lthreads, gThreads);
    }

    //finish
    for (int idx=0; idx<indices.size(); ++idx) {
      int i = indices[idx];
      clFinish(queues[i]);
      //boxm2_block_id id = ids[i];
      //boxm2_opencl_cache* ocl_cache = ocl_caches[i];
      
      //read aux data back into CPU
      //vul_timer ttime; ttime.mark();
      //read_aux(id, ocl_cache, queues[i]);
      //transfer_time += ttime.all();
    }
  }
  float gpu_time = t.all() - transfer_time;

  //unref mems
  for (int i=0; i<queues.size(); ++i) {
    ocl_caches[i]->unref_mem(in_imgs[i].ptr());
  }

//==== DEBUG ====
#if 0
  //unsigned cl_ni=RoundUp(ni,lthreads[0]);
  //unsigned cl_nj=RoundUp(nj,lthreads[1]);
  vil_image_view<float> segLens(cl_ni,cl_nj);
  segLens.fill(0.0f);
  for (int i=0; i<queues.size(); ++i) {
    //clFinish(queues[i]);
    out_imgs[i]->read_to_buffer(queues[i]);
    float* output_arr = (float*) out_imgs[i]->cpu_buffer();
    int count=0;
    for (int j=0; j<cl_nj; ++j)
      for (int i=0; i<cl_ni; ++i)
        segLens(i,j) += output_arr[count++];
  }
  vil_save(segLens, "seg_len.tiff");
#endif
//==============

  //cleanup input image buffer
  if(inImg != img.top_left_ptr())
    delete[] inImg;

  return gpu_time;
}

//: Reads aux memory from GPU to CPU ram
void boxm2_multi_store_aux::read_aux(const boxm2_block_id& id,
                                           boxm2_opencl_cache* opencl_cache,
                                           cl_command_queue&   queue)
{
  //calc data buffer length
  bocl_mem* alpha = opencl_cache->get_data<BOXM2_ALPHA>(id,0,false);
  vcl_size_t alphaTypeSize = boxm2_data_traits<BOXM2_ALPHA>::datasize();
  vcl_size_t dataLen = (vcl_size_t) (alpha->num_bytes() / alphaTypeSize);

  //grab an appropriately sized AUX data buffer
  bocl_mem *aux0 = opencl_cache->get_data<BOXM2_AUX0>(id, dataLen*boxm2_data_traits<BOXM2_AUX0>::datasize());
  bocl_mem *aux1 = opencl_cache->get_data<BOXM2_AUX1>(id, dataLen*boxm2_data_traits<BOXM2_AUX1>::datasize());
  aux0->read_to_buffer(queue);
  aux1->read_to_buffer(queue);
}

//: helper to call ocl kernel - stores aux per block
void boxm2_multi_store_aux::store_aux_per_block(const boxm2_block_id&     id,
                                                      boxm2_scene_sptr    scene,
                                                      boxm2_opencl_cache* opencl_cache,
                                                      cl_command_queue&   queue,
                                                      bocl_kernel*        kernel,
                                                      bocl_mem_sptr&      in_image,
                                                      bocl_mem_sptr&      img_dim,
                                                      bocl_mem_sptr&      ray_o_buff,
                                                      bocl_mem_sptr&      ray_d_buff,
                                                      bocl_mem_sptr&      cl_output,
                                                      bocl_mem_sptr&      lookup,
                                                      vcl_size_t*         lthreads,
                                                      vcl_size_t*         gThreads, 
                                                      bool                store_rgb)
{
  //vcl_cout<<(*id);
  //choose correct render kernel
  boxm2_block_metadata mdata = scene->get_block_metadata(id);
  bocl_kernel* kern =  kernel;

  //write the image values to the buffer
  bocl_mem* blk       = opencl_cache->get_block(id);
  bocl_mem* blk_info  = opencl_cache->loaded_block_info();
  bocl_mem* alpha     = opencl_cache->get_data<BOXM2_ALPHA>(id,0,false);

  //calc data buffer length
  vcl_size_t alphaTypeSize = boxm2_data_traits<BOXM2_ALPHA>::datasize();
  vcl_size_t dataLen = (vcl_size_t) (alpha->num_bytes() / alphaTypeSize);

  //store len in info buffer
  boxm2_scene_info* info_buffer = (boxm2_scene_info*) blk_info->cpu_buffer();
  info_buffer->data_buffer_length = dataLen;
  blk_info->write_to_buffer(queue);

  //grab an appropriately sized AUX data buffer
  int auxTypeSize = (int)boxm2_data_info::datasize(boxm2_data_traits<BOXM2_AUX0>::prefix());
  bocl_mem *aux0   = opencl_cache->get_data<BOXM2_AUX0>(id, dataLen*boxm2_data_traits<BOXM2_AUX0>::datasize());
  bocl_mem *aux1   = opencl_cache->get_data<BOXM2_AUX1>(id, dataLen*boxm2_data_traits<BOXM2_AUX1>::datasize());
  aux0->zero_gpu_buffer(queue);
  aux1->zero_gpu_buffer(queue);

  //set args
  kern->set_arg( blk_info );
  kern->set_arg( blk );
  kern->set_arg( alpha );
  kern->set_arg( aux0 );
  kern->set_arg( aux1 );
  if(store_rgb) {
    bocl_mem *aux2  = opencl_cache->get_data<BOXM2_AUX2>(id, dataLen*boxm2_data_traits<BOXM2_AUX2>::datasize());
    bocl_mem *aux3  = opencl_cache->get_data<BOXM2_AUX3>(id, dataLen*boxm2_data_traits<BOXM2_AUX3>::datasize());
    aux2->zero_gpu_buffer(queue);
    aux3->zero_gpu_buffer(queue);
    kern->set_arg( aux2 ); 
    kern->set_arg( aux3 );
  }
  kern->set_arg( lookup.ptr() );
  kern->set_arg( ray_o_buff.ptr() );
  kern->set_arg( ray_d_buff.ptr() );
  kern->set_arg( img_dim.ptr() );
  kern->set_arg( in_image.ptr() );
  kern->set_arg( cl_output.ptr() );
  kern->set_local_arg( lthreads[0]*lthreads[1]*sizeof(cl_uchar16) );//local tree,
  kern->set_local_arg( lthreads[0]*lthreads[1]*sizeof(cl_uchar4) ); //ray bundle,
  kern->set_local_arg( lthreads[0]*lthreads[1]*sizeof(cl_int) );    //cell pointers,
  kern->set_local_arg( lthreads[0]*lthreads[1]*sizeof(cl_float4) ); //cached aux,
  kern->set_local_arg( lthreads[0]*lthreads[1]*10*sizeof(cl_uchar) ); //cumsum buffer, imindex buffer

  //execute kernel
  kern->execute(queue, 2, lthreads, gThreads);

  //clear render kernel args so it can reset em on next execution
  kern->clear_args();

  //enqueue two nonblocking reads
  //aux0->read_to_buffer(queue, false);
  //aux1->read_to_buffer(queue, false);
}

//-----------------------------------------------------------------
// returns vector of bocl_kernels for this specific device
//-----------------------------------------------------------------
bocl_kernel* boxm2_multi_store_aux::get_kernels(bocl_device_sptr device, vcl_string opts)
{
  // check to see if this device has compiled kernels already
  vcl_string identifier = device->device_identifier()+opts;
  if (kernels_.find(identifier) != kernels_.end())
    return kernels_[identifier];

  //if not, compile and cache them
  vcl_cout<<"===========Compiling multi update kernels===========\n"
          <<"  for device: "<<device->device_identifier()<<vcl_endl;

  //gather all render sources... seems like a lot for rendering...
  vcl_vector<vcl_string> src_paths;
  vcl_string source_dir = boxm2_ocl_util::ocl_src_root();
  src_paths.push_back(source_dir + "scene_info.cl");
  src_paths.push_back(source_dir + "cell_utils.cl");
  src_paths.push_back(source_dir + "bit/bit_tree_library_functions.cl");
  src_paths.push_back(source_dir + "backproject.cl");
  src_paths.push_back(source_dir + "statistics_library_functions.cl");
  src_paths.push_back(source_dir + "ray_bundle_library_opt.cl");
  src_paths.push_back(source_dir + "bit/update_kernels.cl");
  src_paths.push_back(source_dir + "update_functors.cl");
  src_paths.push_back(source_dir + "bit/cast_ray_bit.cl");

  //compilation options
  vcl_string options = opts+" -D INTENSITY  ";
  options += " -D DETERMINISTIC ";

  //create all passes
  bocl_kernel* seg_len = new bocl_kernel();
  vcl_string seg_opts = options + " -D SEGLEN -D STEP_CELL=step_cell_seglen(aux_args,data_ptr,llid,d) ";
  seg_len->create_kernel(&device->context(),device->device_id(), src_paths, "seg_len_main", seg_opts, "update::seg_len");

  //cache in map
  kernels_[identifier] = seg_len;
  return kernels_[identifier];
}


bocl_kernel* boxm2_multi_store_aux::get_kernels_color(bocl_device_sptr device, vcl_string opts)
{
  // compile kernels if not already compiled
  vcl_string identifier = device->device_identifier() + opts + "_color";
  if (kernels_.find(identifier) != kernels_.end())
    return kernels_[identifier];

  //otherwise compile the kernels
  vcl_cout<<"=== boxm2_multi_store_aux::compiling kernels===\n"
          <<"    for device "<<identifier<<vcl_endl;

  vcl_vector<vcl_string> src_paths;
  vcl_string source_dir = boxm2_ocl_util::ocl_src_root();
  src_paths.push_back(source_dir + "scene_info.cl");
  src_paths.push_back(source_dir + "pixel_conversion.cl");
  src_paths.push_back(source_dir + "bit/bit_tree_library_functions.cl");
  src_paths.push_back(source_dir + "backproject.cl");
  src_paths.push_back(source_dir + "statistics_library_functions.cl");
  src_paths.push_back(source_dir + "ray_bundle_library_opt.cl");
  src_paths.push_back(source_dir + "bit/update_rgb_kernels.cl");
  vcl_vector<vcl_string> non_ray_src = vcl_vector<vcl_string>(src_paths);
  src_paths.push_back(source_dir + "update_rgb_functors.cl");
  src_paths.push_back(source_dir + "bit/cast_ray_bit.cl");

  //compilation options
  vcl_string options = " -D INTENSITY ";
  options += " -D YUV -D DETERMINISTIC -D MOG_TYPE_8 ";
  options += opts;

  //seg len pass
  bocl_kernel* seg_len = new bocl_kernel();
  vcl_string seg_opts = options + " -D SEGLEN -D STEP_CELL=step_cell_seglen(aux_args,data_ptr,llid,d) ";
  seg_len->create_kernel(&device->context(), device->device_id(), src_paths, "seg_len_main", seg_opts, "update_color::seg_len");

  //create  compress rgb pass
  bocl_kernel* comp = new bocl_kernel();
  vcl_string comp_opts = options + " -D COMPRESS_RGB ";
  comp->create_kernel(&device->context(), device->device_id(), non_ray_src, "compress_rgb", comp_opts, "update_color::compress_rgb");

  //store and return
  kernels_[identifier] = seg_len;
  return kernels_[identifier];
}
